package com.yanzuoguang.util.thread;

import com.yanzuoguang.util.helper.DateHelper;
import com.yanzuoguang.util.log.Log;

import java.util.*;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

/**
 * 多线程处理队列
 * @author 颜佐光
 */
public abstract class AbstractThreadList<T extends Object> implements ThreadWaitExecute {
    private long threadID;
    /**
     * 线程等待间隔
     */
    public static int PrintTimeout = 1000 * 2;

    /**
     * 线程数量
     */
    private int threadCount;

    /**
     * 线程记录数量
     */
    private int logSize = 1000;

    /**
     * 线程人数
     */
    private volatile ThreadListNum threadListNum = new ThreadListNum();

    /**
     * 缓存队列
     */
    private volatile LinkedBlockingQueue<T> cache = new LinkedBlockingQueue<T>();

    /**
     * 缓存的线程
     */
    private Map<Thread, ThreadListData<T>> cacheThread = new Hashtable<Thread, ThreadListData<T>>();

    /**
     * 缓存的异常数据
     */
    private RuntimeException exception;

    /**
     * 等待对象
     */
    private ThreadWait threadWait = new ThreadWait(this);

    /**
     * 线程开启数量
     *
     * @param threadCount
     */
    public AbstractThreadList(int threadCount) {
        this.threadID = Thread.currentThread().getId();
        threadCount = Math.max(threadCount, 0);
        this.threadCount = threadCount;
    }

    /**
     * 获取线程数量
     *
     * @return
     */
    public int getThreadCount() {
        return threadCount;
    }

    /**
     * 总任务数量
     *
     * @return
     */
    public int getRowTotal() {
        return threadListNum.getRowTotal();
    }

    /**
     * 已经执行的任务数量
     *
     * @return
     */
    public int getRowExecute() {
        return threadListNum.getRowExecute();
    }

    /**
     * 执行错误的任务数量
     *
     * @return
     */
    public int getRowError() {
        return threadListNum.getRowError();
    }

    /**
     * 设置线程数量
     *
     * @param threadCount 线程数量
     * @return
     */
    public void setThreadCount(int threadCount) {
        this.threadCount = threadCount;
    }

    /**
     * 添加需要处理的数据，并自动开启线程,可以反复调用。
     *
     * @param list
     */
    public void add(T... list) {
        if (list != null) {
            this.add(Arrays.asList(list));
        }
    }

    /**
     * 添加需要处理的数据，并自动开启线程,可以反复调用。
     *
     * @param list
     */
    public void add(List<T> list) {
        if (list == null || list.size() == 0) {
            return;
        }
        int size = list.size();
        this.initRowTotal(size);
        for (T item : list) {
            if (item == null) {
                continue;
            }
            this.add(item);
        }
    }

    /**
     * 将数据添加到队列中
     *
     * @param item
     */
    private void add(T item) {
        cache.offer(item);
    }

    /**
     * 初始化任务数量
     *
     * @param total
     */
    private void initRowTotal(int total) {
        threadListNum.init(total);
        this.threadOpen();
    }


    /**
     * 打开线程
     */
    private void threadOpen() {
        if (threadListNum.getRowTotal() < 1) {
            return;
        }
        int threadCount = Math.min(this.threadCount, threadListNum.getRowTotal());
        // 开启剩余线程数量
        for (int i = cacheThread.size(); i < threadCount; i++) {
            // 开启一个新的线程
            ThreadHelper.runThread(new Runnable() {
                @Override
                public void run() {
                    long threadID = Thread.currentThread().getId();
                    try {
                        if (threadID != AbstractThreadList.this.threadID) {
                            Log.threadBegin();
                        }
                        threadExecute();
                    } finally {
                        if (threadID != AbstractThreadList.this.threadID) {
                            Log.threadCommit();
                        }
                    }
                }
            });
            if (i == threadCount - 1 && threadListNum.getRowTotal() > this.logSize) {
                log("已开启线程:" + (i + 1) + " 处理数据:" + threadListNum.getRowTotal());
            }
        }
    }

    /**
     * 判断线程是否还需要等待
     *
     * @return
     */
    private boolean isWait() {
        return cache.size() > 0;
    }

    /**
     * 判断是否已经全部完成
     *
     * @return
     */
    @Override
    public boolean isFinally() {
        if (this.isWait()) {
            return false;
        }
        if (cacheThread.size() == 0) {
            return true;
        }
        return false;
    }

    /**
     * 打印任务状态
     *
     * @description
     */
    @Override
    public void printStatus() {
        int size = cache.size() + cacheThread.size();
        String str = String.format("%s 已处理: %d  错误: %d  剩余: %d 后暂停", DateHelper.getNow(),
                threadListNum.getRowExecute(), threadListNum.getRowError(), size);
        log(str);
    }

    /**
     * 当进度多长时间没有变动时打印状态
     *
     * @return
     */
    @Override
    public int printTimeout() {
        return PrintTimeout;
    }

    /**
     * 获取下一个需要处理的对象
     *
     * @return
     */
    private T getNextItem() {
        try {
            return cache.poll(10, TimeUnit.MILLISECONDS);
        } catch (Exception ex) {
            return null;
        }
    }

    /**
     * 线程执行函数,可以用于主线程执行
     */
    private void threadExecute() {
        cacheThread.put(Thread.currentThread(), new ThreadListData<T>());
        try {
            do {
                next();
                threadWait.nexted();
            } while (this.isWait());
        } finally {
            cacheThread.remove(Thread.currentThread());
            threadWait.finish();
        }
    }

    /**
     * 开始处理下一个对象
     */
    private void next() {
        // 取得需要处理的对象
        T item = getNextItem();

        // 等待下一个对象
        if (item == null) {
            return;
        }

        ThreadListData<T> data = cacheThread.get(Thread.currentThread());
        data.setDate(new Date());
        data.setData(item);
        try {
            this.run(item);
            threadListNum.addSuccess();
        } catch (RuntimeException ex) {
            threadListNum.addError();
            logError(ex);
        }
    }

    /**
     * 开启等待线程，一直等待到全部执行完毕
     */
    public void waitRun() {
        this.threadExecute();
        threadWait.waitFinally();
    }

    /**
     * 关闭所有线程
     */
    public void clear() {
        threadListNum.clear();
        this.cache.clear();
    }

    /**
     * 日志函数
     *
     * @param log
     */
    private void log(String log) {
        Log.info(AbstractThreadList.class, log);
    }

    /**
     * 日志函数
     *
     * @param ex
     */
    private void logError(RuntimeException ex) {
        this.exception = ex;
    }

    /**
     * 抛出线程中的异常数据
     */
    public AbstractThreadList<T> throwThreadError() {
        if (this.exception != null) {
            throw this.exception;
        }
        return this;
    }

    /**
     * 具体的处理执行的函数的数据
     *
     * @param item 需要处理的数据
     */
    public abstract void run(T item);
}
